import pytest

from pydaconf import PydaConf
from pydaconf.utils.exceptions import ProviderException
from pydantic import BaseModel


class Config(BaseModel):
    username: str
    password: str

def test_validation_errors() -> None:
    provider = PydaConf[Config]()
    with pytest.raises(ProviderException):
        provider.from_dict({'username': "test"})
        _ = provider.config



import threading
import time
from collections.abc import Callable
from unittest.mock import Mock, call

import pytest

from pydaconf import ProviderException, PydaConf
from pydaconf.plugins.base import PluginBase
from pydantic import BaseModel


class ConnectionConfig(BaseModel):
    password: str

class User(BaseModel):
    username: str
    password: str

class Config(BaseModel):
    connection: ConnectionConfig
    users: list[User]
    developer: list[str]

class TestUpdatePlugin(PluginBase):
    """Test plugin"""

    PREFIX='ONUPDATE'

    def _run_on_update_in_thread(self, on_update_callback: Callable[[str], None]) -> None:
        def delayed_callback() -> None:
            time.sleep(0.2)
            on_update_callback('Update')

        threading.Thread(target=delayed_callback, daemon=True).start()

    def run(self, value: str, on_update_callback: Callable[[str], None]) -> str:
        self._run_on_update_in_thread(on_update_callback)
        return value


def test_on_update() -> None:
    mock_callback_connection_password = Mock()
    mock_callback_users_password = Mock()
    config = {
        "connection": {
            "password": "ONUPDATE:///Test1"  
        },
        "users": [
            { "username": "test", "password": "ONUPDATE:///Test2"},
            { "username": "test2", "password": "ONUPDATE:///Test3"},
        ],
        "developer": ["ONUPDATE:///Test4"]
    }
    provider = PydaConf[Config]()
    provider.register_plugin(TestUpdatePlugin)
    provider.from_dict(config)
    provider.on_update(r'.connection.password', mock_callback_connection_password)
    provider.on_update(r'.users.+', mock_callback_users_password)
    assert provider.config
    time.sleep(1)

    assert mock_callback_connection_password.call_count == 1
    expected_calls = [call('.connection.password', 'Update')]
    mock_callback_connection_password.assert_has_calls(expected_calls)


    assert mock_callback_users_password.call_count == 2
    expected_calls = [
        call('.users[0].password', 'Update'),
        call('.users[1].password', 'Update')]
    mock_callback_users_password.assert_has_calls(expected_calls, any_order=True)


def test_on_update_not_initialized() -> None:
    with pytest.raises(ProviderException):
        provider = PydaConf[Config]()
        provider._on_update('test', 'test')



import os
import sys

from pydaconf.utils.exceptions import UnknownFormat


def guess_content_format(content: str) -> str | None:
    """Guess file format based on the first line. This is much faster for big config files than using the parsers."""

    first_line = next(iter(content.splitlines()), None)
    if first_line is None:
        raise UnknownFormat("Config file first line is empty. Unable to identify the format")

    if first_line.startswith("{") or (first_line.startswith("[") and "]" in first_line and "=" not in first_line):
        return 'json'

    elif ":" in first_line and not first_line.startswith("["):
        return 'yaml'

    elif first_line.startswith("[") and "=" not in first_line:
        return 'toml'

    elif "=" in first_line:
        return 'toml'

    else:
        return None

def load_yaml_content(content: str) -> dict:
    try:
        import yaml
    except ImportError as e:
        raise ImportError('PyYAML is not installed, run `pip install pydaconf[yaml]`') from e

    result: dict | None = yaml.load(content, Loader=yaml.FullLoader)
    if result is None:
        raise ValueError("Empty yaml content")

    return result
    

def load_json_content(content: str) -> dict:
    import json
    result: dict | None = json.loads(content)
    if result is None:
        raise ValueError("Empty Json content")

    return result
    

def load_toml_content(content: str) -> dict:
    if sys.version_info < (3, 11):
        try:
            import tomli
            toml_module = tomli
        except ImportError as e:
            raise ImportError('Toml is not installed, run `pip install pydaconf[toml]`') from e
    else:
        import tomllib
        toml_module = tomllib
    
    return toml_module.loads(content)

def load_config_file(file_path: str) -> dict:
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"Configuration file '{file_path}' doesn't exist")
    
    with open(file_path) as f:
        content = f.read()

    # Guess the file based on extension
    _, extension = os.path.splitext(file_path)
    if extension.lower() in {'.yaml', '.yml'}:
        return load_yaml_content(content)

    elif extension.lower() == '.json':
       return load_json_content(content)

    elif extension.lower() == '.toml':
        return load_toml_content(content)
    else:
        discovered_format = guess_content_format(content)
        if discovered_format == 'json':
            return load_json_content(content)
        elif discovered_format == 'yaml':
            return load_yaml_content(content)
        elif discovered_format == 'toml':
            return load_toml_content(content)
        else:
            raise UnknownFormat("Unknown config file format. Valid formats are json, yaml and toml.")

    
def load_from_url(url: str, headers: dict | None = None) -> dict:
    try:
        import requests
    except ImportError as e:
        raise ImportError('requests is not installed, run `pip install pydaconf[requests]`') from e
    
    response = requests.get(url, headers=headers)
    response.raise_for_status()
    content = response.text
    discovered_format = guess_content_format(content)
    if discovered_format == 'json':
        return load_json_content(content)
    elif discovered_format == 'yaml':
        return load_yaml_content(content)
    elif discovered_format == 'toml':
        return load_toml_content(content)
    else:
        raise UnknownFormat("Unknown config file format")

    


from collections.abc import Callable
from unittest.mock import MagicMock, patch

from pydaconf import PydaConf
from pydaconf.plugins.base import PluginBase
from pydaconf.utils.plugins import load_builtin_plugins, load_dynamic_plugins
from pydantic import BaseModel


class InvalidPlugin:
    pass

class DummyPlugin(PluginBase):

    PREFIX = 'DUMMY'

    def run(self, value: str, on_update_callback: Callable[[str], None]) -> str:
        return value


def test_load_builtin_plugins() -> None:
    plugins = list(load_builtin_plugins())
    assert len(plugins) > 0


def test_load_dynamic_plugins() -> None:
    with patch("importlib.metadata.entry_points") as mock_entry_points:
        dummy_plugin = MagicMock()
        dummy_plugin.load.return_value = DummyPlugin

        mock_entry_points.return_value = [dummy_plugin]

        # Execute the function under test
        plugins = list(load_dynamic_plugins())

        # Assertions
        assert len(plugins) == 1
        assert isinstance(plugins[0], DummyPlugin)
        assert plugins[0].run('test', lambda d: None) == 'test'

def test_skip_invalid_plugin() -> None:
    with patch("importlib.metadata.entry_points") as mock_entry_points:
        invalid_plugin = MagicMock()
        invalid_plugin.load.return_value = InvalidPlugin

        mock_entry_points.return_value = [invalid_plugin]

        # Execute the function under test
        plugins = list(load_dynamic_plugins())

        # Assertions
        assert len(plugins) == 0


def test_provider_load_dynamic_plugins() -> None:
    class Config(BaseModel):
        test: str

    with patch("importlib.metadata.entry_points") as mock_entry_points:
        dummy_plugin = MagicMock()
        dummy_plugin.load.return_value = DummyPlugin

        mock_entry_points.return_value = [dummy_plugin]

        provider = PydaConf[Config]()
        provider.from_dict({'test': 'test'})
        _ = provider.config
        assert isinstance(provider._plugins['DUMMY'], DummyPlugin)



from unittest.mock import MagicMock, patch

from pydaconf import PydaConf
from pydantic import BaseModel


class Config(BaseModel):
    name: str
    secret: str

def test_env_plugin() -> None:
    config = {
        "name": "Test",
        "secret": "ENV:///TEST"
    }
    conf_provider = PydaConf[Config]()
    with patch("os.environ") as environ:
        environ.get = MagicMock(return_value = 'TEST')

        conf_provider.from_dict(config)
        assert conf_provider.config.secret == 'TEST'


try:
    import typer
    from cryptography.fernet import Fernet
except ImportError as e:
    raise ImportError('CLI dependencies are not installed, run `pip install pydaconf[cli]`') from e

app = typer.Typer()
seal = typer.Typer()
app.add_typer(seal, name="seal")

# TODO: Add better error handling

@seal.command('generate-key')
def generate_key() -> None:
    """Generate symmetric encryption key for sealed secrets"""
    typer.echo(Fernet.generate_key().decode())

@seal.command('encrypt')
def generate(key: str, secret: str) -> None:
    """Encrypt secret with symmetric encryption key"""
    f = Fernet(key)
    typer.echo(f.encrypt(secret.encode()).decode())

@seal.command('decrypt')
def decrypt(key: str, secret: str) -> None:
    """Encrypt secret with symmetric encryption key"""
    f = Fernet(key)
    typer.echo(f.decrypt(secret).decode())


if __name__ == "__main__":
    app()


import os
from collections.abc import Callable

from pydaconf.plugins.base import PluginBase
from pydaconf.utils.exceptions import ProviderException


class SealedPlugin(PluginBase):
    """Load sealed secret. Load/Decrypt sealed secret as variable based on prefix SEALED:///"""

    PREFIX='SEALED'

    def __init__(self) -> None:
        self.cache: dict[str, str] = {}

    def run(self, value: str, on_update_callback: Callable[[str], None]) -> str:
        if value not in self.cache:
            key = os.environ.get('PYDACONF_SEALED_KEY')
            if key is None:
                raise ProviderException(f"PYDACONF_SEALED_KEY environment variable is required to decrypt '{value}'")
            
            try:
                from cryptography.fernet import Fernet
                encryption_provider = Fernet(key)
            except ImportError as e:
                raise ImportError('Cryptography is not installed, run `pip install pydaconf[cryptography]`') from e

            content = encryption_provider.decrypt(value)

            self.cache[value] = content.decode()
    
        return self.cache[value]




from collections.abc import Callable

from pydaconf.plugins.base import PluginBase


class FileContentPlugin(PluginBase):
    """Load file content as variable plugin. Loads file content as variable based on prefix FILE_CONTENT:///"""

    PREFIX='FILE_CONTENT'

    def __init__(self) -> None:
        self.cache: dict[str, str]= {}

    def run(self, value: str, on_update_callback: Callable[[str], None]) -> str:
        if value not in self.cache:
            try:
                with open(value) as file:
                    content = file.read()

            except Exception as e:
                raise Exception(f"Secret cannot be loaded from file '{value}'. Error", e)

            self.cache[value] = content
    
        return self.cache[value]



from abc import ABC, abstractmethod
from collections.abc import Callable

from pydaconf.utils.exceptions import PluginException


class PluginBase(ABC):
    """Abstract Base Class for plugins."""

    def _execute_plugin(self, value: str, on_update_callback: Callable[[str], None]) -> str:
        """This method is intended to be called only from PyDaConf provider"""
        try:
            return self.run(value, on_update_callback)
        except Exception as e:
            raise PluginException(
                f"Plugin '{self.__class__.__name__}' failed to get data for value '{value}'. Error: {str(e)}") from e

    @property
    @abstractmethod
    def PREFIX(self) -> str:
        raise NotImplementedError


    @abstractmethod
    def run(self, value: str, on_update_callback: Callable[[str], None]) -> str:
        raise NotImplementedError


import pytest

from pydaconf import PydaConf
from pydaconf.utils.exceptions import ProviderException
from pydantic import BaseModel


class Config(BaseModel):
    username: str
    password: str

def test_not_initialized() -> None:
    provider = PydaConf[Config]()
    with pytest.raises(ProviderException):
        assert provider.config


def test_second_initialized() -> None:
    provider = PydaConf[Config]()
    provider.from_dict({'username': 'test', 'password': 'pass'})
    assert provider.config.username == 'test'
    assert provider.config.password == 'pass'



from pydaconf import PydaConf
from pydantic import BaseModel


class Config(BaseModel):
    name: str
    secret: int

def test_file_content_load() -> None:
    config = {
       "name": "Test",
       "secret": "FILE_CONTENT:///tests/resources/file_content_plugin.txt"
    }
    provider = PydaConf[Config]()
    provider.from_dict(config)
    assert provider.config.secret == 24


from unittest.mock import patch

import pytest

from pydaconf import PydaConf
from pydaconf.utils.exceptions import PluginException
from pydantic import BaseModel


class Config(BaseModel):
    secret: str

def test_sealed_exception() -> None:
    config = {
        "secret": "SEALED:///gAAAAABnpRqHdho_dFB7iMs_Ufv5nu59LwqQfE3YCjaDTyg-YrYsEpK8bfvdKsdrj6kCxrAjezCGAdVM0FhzwyfYFIgqnquw8w=="
    }

    with patch('os.environ.get', return_value='WrongKey'):
        with pytest.raises(PluginException):
            provider: PydaConf[Config] = PydaConf[Config]()
            provider.from_dict(config)
            _ = provider.config

def test_sealed_load()  -> None:
    config = {
       "secret": "SEALED:///gAAAAABnpRqHdho_dFB7iMs_Ufv5nu59LwqQfE3YCjaDTyg-YrYsEpK8bfvdKsdrj6kCxrAjezCGAdVM0FhzwyfYFIgqnquw8w=="
    }

    with patch('os.environ.get', return_value='ky8Uv_xkl4WgvHjrrdffJDRV1D5G5pqBTBD0TNGtAi0='):
        provider: PydaConf[Config] = PydaConf[Config]()
        provider.from_dict(config)
        loaded_config = provider.config
        assert loaded_config.secret == 'Test Secret'


import importlib
import importlib.metadata
import inspect
import os
import pkgutil
from collections.abc import Iterator

import pydaconf.plugins
from pydaconf.plugins.base import PluginBase


def load_builtin_plugins() -> Iterator[PluginBase]:
    for _, module_name, _ in pkgutil.iter_modules([str(f"{os.path.dirname(pydaconf.plugins.base.__file__)}")]):
        full_module_name = f"{pydaconf.plugins.__name__}.{module_name}"
        module = importlib.import_module(full_module_name) 

        # Inspect module and find classes that subclass PluginBase
        for _, plugin_class in inspect.getmembers(module, inspect.isclass):
            if issubclass(plugin_class, PluginBase) and plugin_class is not PluginBase:
                yield plugin_class()

def load_dynamic_plugins() -> Iterator[PluginBase]:
    for entry_point in importlib.metadata.entry_points(group='pydaconf.plugins'):
        plugin_class = entry_point.load()
        if issubclass(plugin_class, PluginBase):
            yield plugin_class()


from collections.abc import Callable

from pydaconf import PydaConf
from pydaconf.plugins.base import PluginBase
from pydantic import BaseModel


class Config(BaseModel):
    name: str
    secret: str

class TestPlugin(PluginBase):

    PREFIX = "TEST"

    def run(self, value: str, on_update_callback: Callable) -> str:
        return value
    

def test_register_plugin() -> None:
    config = {
       "name": 'test',
       "secret": "TEST:///TEST"
    }
    provider = PydaConf[Config]()
    provider.from_dict(config)
    provider.register_plugin(TestPlugin)
    assert provider.config.secret == 'TEST'


from pydaconf import PydaConf
from pydantic import BaseModel


class BaseConfig(BaseModel):
    username: str

class SecureConfig(BaseConfig):
    password: str

def test_class_inheritance() -> None:
    config = {
        'username': 'test',
        'password': 'test'
    }

    provider = PydaConf[SecureConfig]()
    provider.from_dict(config)
    assert str(provider) == "PydaConf({'username': 'test', 'password': 'test'})"


import pytest

from pydaconf import ProviderException, PydaConf

config = {
    'username': 'test',
    'password': 'test'
}

def test_missing_base_model() -> None:
    with pytest.raises(ProviderException, match=r"PydaConf must be defined as generic Config\[MyPydanticType\]()"):
        provider = PydaConf() # type: ignore
        provider.from_dict(config)
        _ = provider.config

def test_wrong_type() -> None:
    class Config(dict):
        username: str
        password: str

    with pytest.raises(ProviderException, match=r"Generic type must inherit pydantic BaseModel class Config\[MyPydanticType\]()"):
        provider = PydaConf[Config]() # type: ignore
        provider.from_dict(config)
        _ = provider.config


import os
from unittest.mock import MagicMock, patch

from pydaconf import PydaConf
from pydaconf.utils.file import load_config_file
from pydantic import BaseModel


class Config(BaseModel):
    name: str

def test_provider_json() -> None:
    provider = PydaConf[Config]()
    provider.from_file(os.path.join('tests', 'resources', 'config.json'))
    assert provider.config.name == 'Test'

def test_provider_yaml() -> None:
    provider = PydaConf[Config]()
    provider.from_file(os.path.join('tests', 'resources', 'config.yaml'))
    assert provider.config.name == 'Test'

def test_provider_toml() -> None:
    provider = PydaConf[Config]()
    provider.from_file(os.path.join('tests', 'resources', 'config.toml'))
    assert provider.config.name == 'Test'

def test_load_from_url() -> None:
    mock_response = MagicMock()
    mock_response.text = """name: Test"""
    
    with patch("requests.get", return_value=mock_response):
        provider = PydaConf[Config]()
        provider.from_url('http://dummy.url/config.yaml')
        assert provider.config.name == 'Test'

def test_yaml_content_guess() -> None:
    config = load_config_file(os.path.join('tests', 'resources', 'config.conf-yaml'))
    assert config['name'] == 'Test'

def test_json_content_guess() -> None:
    config = load_config_file(os.path.join('tests', 'resources', 'config.conf-json'))
    assert config['name'] == 'Test'

def test_toml_content_guess() -> None:
    config = load_config_file(os.path.join('tests', 'resources', 'config.conf-toml'))
    assert config['name'] == 'Test'




def get_dict_value(data: dict, key_path: str) -> str | None | dict:
    keys = key_path.split('.')
    for key in keys:
        if isinstance(data, dict) and key in data:
            data = data[key]
        else:
            return None
    
    return data



import os
from collections.abc import Callable

from dotenv import load_dotenv

from pydaconf.plugins.base import PluginBase

# TODO: Remove the cache since the provider implements its own cache

class EnvPlugin(PluginBase):
    """Environment variables plugin. Loads environment variable based on prefix ENV:///"""

    PREFIX='ENV'

    def __init__(self) -> None:
        self.cache: dict[str, str] = {}

    def run(self, value: str, on_update_callback: Callable[[str], None]) -> str:
        load_dotenv()
        if value not in self.cache:
            env_var = os.environ.get(value)
            if not env_var:
                raise Exception(f"Secret cannot be loaded for '{value}'. Environment '{value}' is empty.")
            
            self.cache[value] = env_var
    
        return self.cache[value]



import pytest

from pydaconf import ProviderException, PydaConf
from pydantic import BaseModel


class Config(BaseModel):
    name: str
    secret: str

def test_env_plugin() -> None:
    config = {
        "name": "Test",
        "secret": "WRONG:///TEST"
    }
    conf_provider = PydaConf[Config]()
    with pytest.raises(ProviderException, match="Plugin with prefix 'WRONG' is not registered"):
        conf_provider.from_dict(config)
        _ = conf_provider.config


import pytest

from pydaconf import PydaConf
from pydaconf.utils.exceptions import InterpolationException
from pydantic import BaseModel


class Parameters(BaseModel):
    pool: bool
    replicas: dict[str, str]


class Connection(BaseModel):
    protocol: str
    port: int
    host: str
    parameters: Parameters


class Config(BaseModel):
    domain: str
    users: list[str]
    connection: Connection
    url: str


def test_interpolation_with_spaces()  -> None:
    config = {
        "domain": "example.com",
        "users": ["admin@{{ domain }}"],
        "connection": {
            "protocol": "https",
            "host": "example.com",
            "port": 443,
            "parameters": {
                "pool": True,
                "replicas": {
                    "primary": "primary@{{ domain }}",
                    "secondary": "secondary@{{ domain }}"
                }
            }

        },
        "url": "{{ connection.protocol }}://{{ connection.host }}:{{ connection.port }}"
    }

    provider = PydaConf[Config]()
    provider.from_dict(config)
    c = provider.config
    assert c.url == "https://example.com:443"
    assert c.connection.parameters.replicas['primary'] == "primary@example.com"
    assert c.connection.parameters.replicas['secondary'] == "secondary@example.com"
    assert c.users == ["admin@example.com"]

def test_interpolation_without_spaces()  -> None:
    config = {
        "domain": "example.com",
        "users": ["admin@{{domain}}"],
        "connection": {
            "protocol": "https",
            "host": "example.com",
            "port": 443,
            "parameters": {
                "pool": True,
                "replicas": {
                    "primary": "primary@{{domain}}",
                    "secondary": "secondary@{{domain}}"
                }
            }
        },
        "url": "{{connection.protocol}}://{{connection.host}}:{{connection.port}}"
    }

    provider: PydaConf[Config] = PydaConf[Config]()
    provider.from_dict(config)
    c = provider.config
    assert c.url == "https://example.com:443"
    assert c.connection.parameters.replicas['primary'] == "primary@example.com"
    assert c.connection.parameters.replicas['secondary'] == "secondary@example.com"
    assert c.users == ["admin@example.com"]

def test_interpolation_raise_exception() -> None:
    config = {
        "domain": "example.com",
        "users": ["admin@{{ domain }}"],
        "connection": {
            "protocol": "https",
            "host": "example.com",
            "port": 443,
            "parameters": {
                "pool": True,
                "replicas": {
                    "primary": "primary@{{ domain }}",
                    "secondary": "secondary@{{ domain }}"
                }
            }

        },
        # connection.schema is not defined
        "url": "{{connection.schema}}://{{connection.host}}:{{connection.port}}"
    }

    provider = PydaConf[Config]()
    with pytest.raises(InterpolationException):
        provider.from_dict(config)
        assert provider.config


import logging
import re
import threading
from collections.abc import Callable
from functools import partial
from typing import Generic, TypeAlias, TypeVar, get_args

from pydaconf.plugins.base import PluginBase
from pydaconf.utils.exceptions import ProviderException
from pydaconf.utils.file import load_config_file, load_from_url
from pydaconf.utils.interpolation import (
    has_interpolation_template,
    interpolate_template,
)
from pydaconf.utils.plugins import load_builtin_plugins, load_dynamic_plugins
from pydantic import BaseModel, ValidationError

T = TypeVar("T", bound=BaseModel)
ConfigValueType: TypeAlias = list | dict | str | int | bool | None

class PydaConf(Generic[T]):

    def __init__(self) -> None:
        self._raw_config: dict | None = None
        self._config: T | None = None
        self._plugins: dict[str, PluginBase] = {}
        self._update_subscribers: dict[str, list[Callable[[str, str], None]]] = {}
        self._update_lock = threading.Lock()
        self.logger = logging.getLogger(__name__)

    def from_file(self, file_path: str) -> None:
        self._load_plugins()
        self._raw_config = load_config_file(file_path)

    def from_url(self, url: str) -> None:
        self._load_plugins()
        self._raw_config = load_from_url(url)

    def from_dict(self, dict_data: dict) -> None:
        self._load_plugins()
        self._raw_config = dict_data

    @property
    def config(self) -> T:
        if self._raw_config is None:
            raise ProviderException("""PydaConf is not initialized. 
                                    You need to run on of these methods first from_file('file_path'), from_dict(dict_data) or from_url(url).""")

        if self._config is None:
            config_copy = self._raw_config.copy()
            self.logger.debug("Inject credentials to the config")
            self._inject_secrets(config_copy)
            self.logger.debug("Interpolate values")
            self._interpolate_templates(config_copy, config_copy)

            try:
                self.logger.debug("Build config object")
                config: T = self._get_generic_type()(**config_copy)
                self._config = config
            except ValidationError as e:
                raise ProviderException('Configuration file validation failed with errors', e.errors()) from e

        return self._config

    def register_plugin(self, plugin_class: type[PluginBase]) -> None:
        """ Manually register plugin """
        self._plugins[str(plugin_class.PREFIX)] = plugin_class()

    def on_update(self, key_pattern: str, callback: Callable[[str, str], None]) -> None:
        """Subscribe to an update events of specific pattern."""
        self._update_subscribers.setdefault(key_pattern, [])
        self._update_subscribers[key_pattern].append(callback)

    def _get_generic_type(self) -> type[T]:
        """Return the type of generic T """

        # This is a bit hacky method since __orig_class__ is not well documented and could be changed in future...
        orig_class = getattr(self, '__orig_class__', None)
        if orig_class is None:
            raise ProviderException('PydaConf must be defined as generic Config[MyPydanticType]()')

        generic_type: type[T] | None = next(iter(get_args(orig_class)), None)
        if generic_type is None or not issubclass(generic_type, BaseModel):
            raise ProviderException('Generic type must inherit pydantic BaseModel class Config[MyPydanticType]()')

        return generic_type

    def _load_plugins(self) -> None:
        self._load_builtin_plugins()
        self._load_dynamic_plugins()
        
    def _load_builtin_plugins(self) -> None:
        for plugin in load_builtin_plugins():
            self._plugins[plugin.PREFIX] = plugin
    
    def _load_dynamic_plugins(self) -> None:
        for plugin in load_dynamic_plugins():
            self._plugins[plugin.PREFIX] = plugin

    def _interpolate_templates(self, node: ConfigValueType, config_data: dict) -> None:
        if isinstance(node, list):
            for index, element in enumerate(node):
                if type(element) is str:
                    node[index] = interpolate_template(element, config_data)
                else:
                    self._interpolate_templates(element, config_data)

        elif isinstance(node, dict):
            for key, element in node.items():
                if isinstance(element, dict) or isinstance(element, list):
                    self._interpolate_templates(element, config_data)
                
                elif isinstance(element, str):
                    if has_interpolation_template(element):
                        node[key] = interpolate_template(element, config_data)


    def _match_and_execute_plugin(self, element: str, key: str) -> str:
        match_prefix = re.match(r'(?P<PLUGIN_PREFIX>[^:]+):///(?P<VALUE>[^\s]+)', element)
        if match_prefix:
            plugin_prefix = match_prefix.groupdict()['PLUGIN_PREFIX']
            value = match_prefix.groupdict()['VALUE']
            plugin = self._plugins.get(plugin_prefix.upper())
            if plugin is None:
                raise ProviderException(f"Plugin with prefix '{plugin_prefix}' is not registered")
            return plugin._execute_plugin(value, partial(self._on_update, key))
        else:
            return element

    def _inject_secrets(self, node: ConfigValueType, key_path: str="") -> None:
        if isinstance(node, list):
            for index, element in enumerate(node):
                if type(element) is str:
                   node[index] = self._match_and_execute_plugin(element, f"{key_path}[{index}]")
                else:
                    self._inject_secrets(element, key_path=f"{key_path}[{index}]")

        elif isinstance(node, dict):
            for key, element in node.items():
                if isinstance(element, dict) or isinstance(element, list):
                    self._inject_secrets(element, key_path=f"{key_path}.{key}")

                elif isinstance(element, str):
                    node[key] = self._match_and_execute_plugin(element, f"{key_path}.{key}")

    def _update_config(self, key: str, value: str) -> None:
        """ Update the configuration base on key and value """

        # Unfortunately, I could find a better way to do this directly on the pyndatic model
        # TODO: Research a better option with setattr and getattr
        if self._config is None:
            raise ProviderException('PydaConf is not initialized, or you call on_update callback in the plugin run.')

        config_model = self._config.model_dump()
        current = config_model
        keys = re.split(r'\.(?![^\[]*\])', key.lstrip('.'))  # Split while ignoring dots inside brackets

        for key in keys[:-1]:  # Traverse until the second last key
            match = re.match(r'(\w+)\[(\d+)\]', key)  # Match list indexing pattern, e.g., users[0]

            if match:
                key, index = match.groups()
                index = int(index)
                current = current[key][index]  # Move to the specific index
            else:
                current = current[key]  # Move deeper

        # Process the last key
        final_key = keys[-1]
        match = re.match(r'(\w+)\[(\d+)\]', final_key)

        if match:
            key, index = match.groups()
            index = int(index)
            current[key][index] = value  # Update value at the index
        else:
            current[final_key] = value  # Update the final key

        self._config = self._get_generic_type()(**config_model)


    def _on_update(self, key: str, value: str) -> None:
        """ Update the configuration and notify all subscribers registered for specific key pattern """

        # We use thread lock to protection against race conditions when threads access shared objects.
        with self._update_lock:
            self._update_config(key, value)

            for key_patters, subscribers in self._update_subscribers.items():
                if re.match(key_patters, key):
                    for subscriber in subscribers:
                        subscriber(key, value)


    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.config.model_dump()})'



import re

from pydaconf.utils.dict import get_dict_value
from pydaconf.utils.exceptions import InterpolationException


def has_interpolation_template(value: str) -> bool:
    """Check if a string contains placeholders for interpolation."""

    return bool(re.search(r"{{\s*(.*?)\s*}}", value))

def interpolate_template(value: str, data: dict) -> str:
    """Replace {{ ... }} placeholders with values from the data dictionary."""
    
    def replace_match(match: re.Match) -> str:
        key_path = match.group(1).strip()  # Extract the key path
        match_value = get_dict_value(data, str(key_path))
        if match_value is None:
            raise InterpolationException(f"Interpolation failed for key '{key_path}'")
        
        return str(match_value)

    return re.sub(r"{{\s*(.*?)\s*}}", replace_match, value)